'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var typeOf = require('component-type');
var _typecast = require('typecast');

/**
 * A property instance gets returned whenever you call `schema.path()`.
 * Properties are also created internally when an object is passed to the Schema constructor.
 *
 * @param {String} name - the name of the property
 * @param {Schema} schema - parent schema
 */

var Property = function () {
  function Property(name, schema) {
    _classCallCheck(this, Property);

    this.name = name;
    this.registry = {};
    this._schema = schema;
    this._type = null;
    this.messages = schema.messages;
    this.validators = schema.validators;
  }

  /**
   * Mount given `schema` on current path.
   *
   * @example
   * const user = new Schema({ email: 'string' })
   * prop.schema(user)
   *
   * @param {Schema} schema - the schema to mount
   * @return {Property}
   */

  _createClass(Property, [{
    key: 'schema',
    value: function schema(_schema) {
      this._schema.path(this.name, _schema);
      return this;
    }

    /**
     * Validate using named functions from the given object.
     * Error messages can be defined by providing an object with
     * named error messages/generators to `schema.message()`
     *
     * The message generator receives the value being validated,
     * the object it belongs to and any additional arguments.
     *
     * @example
     * const schema = new Schema()
     * const prop = schema.path('some.path')
     *
     * schema.message({
     *   binary: (path, ctx) => `${path} must be binary.`,
     *   bits: (path, ctx, bits) => `${path} must be ${bits}-bit`
     * })
     *
     * prop.use({
     *   binary: (val, ctx) => /^[01]+$/i.test(val),
     *   bits: [(val, ctx, bits) => val.length == bits, 32]
     * })
     *
     * @param {Object} fns - object with named validation functions to call
     * @return {Property}
     */

  }, {
    key: 'use',
    value: function use(fns) {
      var _this = this;

      Object.keys(fns).forEach(function (name) {
        var arr = fns[name];
        if (!Array.isArray(arr)) arr = [arr];
        var fn = arr.shift();
        _this._register(name, arr, fn);
      });

      return this;
    }

    /**
     * Registers a validator that checks for presence.
     *
     * @example
     * prop.required()
     *
     * @param {Boolean} [bool] - `true` if required, `false` otherwise
     * @return {Property}
     */

  }, {
    key: 'required',
    value: function required() {
      var bool = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : true;

      return this._register('required', [bool]);
    }

    /**
     * Registers a validator that checks if a value is of a given `type`
     *
     * @example
     * prop.type('string')
     *
     * @param {String} type - type to check for
     * @return {Property}
     */

  }, {
    key: 'type',
    value: function type(_type) {
      this._type = _type;
      return this._register('type', [_type]);
    }

    /**
     * Registers a validator that checks length.
     *
     * @example
     * prop.length({ min: 8, max: 255 })
     *
     * @param {Object} rules - object with `.min` and `.max` properties
     * @param {Number} rules.min - minimum length
     * @param {Number} rules.max - maximum length
     * @return {Property}
     */

  }, {
    key: 'length',
    value: function length(rules) {
      return this._register('length', [rules]);
    }

    /**
     * Registers a validator for enums.
     *
     * @example
     * prop.enum(['cat', 'dog'])
     *
     * @param {Array} rules - allowed values
     * @return {Property}
     */

  }, {
    key: 'enum',
    value: function _enum(enums) {
      return this._register('enum', [enums]);
    }

    /**
     * Registers a validator that checks if a value matches given `regexp`.
     *
     * @example
     * prop.match(/some\sregular\sexpression/)
     *
     * @param {RegExp} regexp - regular expression to match
     * @return {Property}
     */

  }, {
    key: 'match',
    value: function match(regexp) {
      return this._register('match', [regexp]);
    }

    /**
     * Registers a validator that checks each value in an array against given `rules`.
     *
     * @example
     * prop.each({ type: 'string' })
     * prop.each([{ type: 'number' }])
     * prop.each({ things: [{ type: 'string' }]})
     * prop.each(schema)
     *
     * @param {Array|Object|Schema|Property} rules - rules to use
     * @return {Property}
     */

  }, {
    key: 'each',
    value: function each(rules) {
      this._schema.path(this.name + '.$', rules);
      return this;
    }

    /**
     * Proxy method for schema path. Makes chaining properties together easier.
     *
     * @example
     * schema
     *   .path('name')
     *     .type('string')
     *     .required()
     *   .path('email')
     *     .type('string')
     *     .required()
     *
     */

  }, {
    key: 'path',
    value: function path() {
      var _schema2;

      return (_schema2 = this._schema).path.apply(_schema2, arguments);
    }

    /**
     * Typecast given `value`
     *
     * @example
     * prop.type('string')
     * prop.typecast(123) // => '123'
     *
     * @param {Mixed} value - value to typecast
     * @return {Mixed}
     */

  }, {
    key: 'typecast',
    value: function typecast(value) {
      if (!this._type) return value;
      return _typecast(value, this._type);
    }

    /**
     * Validate given `value`
     *
     * @example
     * prop.type('number')
     * assert(prop.validate(2) == false)
     * assert(prop.validate('hello world') instanceof Error)
     *
     * @param {Mixed} value - value to validate
     * @param {Object} ctx - the object containing the value
     * @param {String} [path] - path of the value being validated
     * @return {Error|Boolean}
     */

  }, {
    key: 'validate',
    value: function validate(value, ctx) {
      var path = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this.name;

      var types = Object.keys(this.registry);
      var done = {};
      var error = void 0;

      // Required first
      if (error = this._run('required', value, ctx, path)) {
        return error;
      }

      // No need to continue if value is null-ish
      if (value == null) return false;

      // Run type second
      if (error = this._run('type', value, ctx, path)) {
        return error;
      }

      // Make sure required and run are not executed again
      done.required = true;
      done.type = true;

      // Run the rest
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = types[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var type = _step.value;

          if (done[type]) continue;
          if (error = this._run(type, value, ctx, path)) {
            return error;
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }

      return false;
    }

    /**
     * Run validator of given `type`
     *
     * @param {String} type - type of validator
     * @param {Mixed} value - value to validate
     * @param {Object} ctx - the object containing the value
     * @param {String} path - path of the value being validated
     * @return {Error}
     * @private
     */

  }, {
    key: '_run',
    value: function _run(type, value, ctx, path) {
      if (!this.registry[type]) return;
      var _registry$type = this.registry[type],
          args = _registry$type.args,
          fn = _registry$type.fn;

      var validator = fn || this.validators[type];
      var valid = validator.apply(undefined, [value, ctx].concat(_toConsumableArray(args)));
      if (!valid) return this._error(type, ctx, args, path);
    }

    /**
     * Register validator
     *
     * @param {String} type - type of validator
     * @param {Array} args - argument to pass to validator
     * @param {Function} [fn] - custom validation function to call
     * @return {Property}
     * @private
     */

  }, {
    key: '_register',
    value: function _register(type, args, fn) {
      this.registry[type] = { args: args, fn: fn };
      return this;
    }

    /**
     * Create an error
     *
     * @param {String} type - type of validator
     * @param {Object} ctx - the object containing the value
     * @param {Array} args - arguments to pass
     * @param {String} path - path of the value being validated
     * @return {Error}
     * @private
     */

  }, {
    key: '_error',
    value: function _error(type, ctx, args, path) {
      var message = this.messages[type] || this.messages.default;

      if (typeof message == 'function') {
        message = message.apply(undefined, [path, ctx].concat(_toConsumableArray(args)));
      }

      var err = new Error(message);
      err.path = path;
      return err;
    }
  }]);

  return Property;
}();

module.exports = Property;