'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Property = require('./property');
var typeOf = require('component-type');
var dot = require('eivindfjeldstad-dot');

var Messages = require('./messages');
var Validators = require('./validators');

/**
 * A Schema defines the structure that objects should be validated against.
 *
 * @example
 * const post = new Schema({
 *   title: {
 *     type: 'string',
 *     required: true,
 *     length: { min: 1, max: 255 }
 *   },
 *   content: {
 *     type: 'string',
 *     required: true
 *   },
 *   published: {
 *     type: 'date',
 *     required: true
 *   },
 *   keywords: [{ type: 'string' }]
 * })
 *
 * @example
 * const author = new Schema({
 *   name: {
 *     type: 'string',
 *     required: true
 *   },
 *   email: {
 *     type: 'string',
 *     required: true
 *   },
 *   posts: [post]
 * })
 *
 * @param {Object} [obj] - schema definition
 * @param {Object} [opts] - options
 * @param {Boolean} [opts.typecast=false] - typecast values before validation
 * @param {Boolean} [opts.strip=true] - strip properties not defined in the schema
 */

var Schema = function () {
  function Schema() {
    var _this = this;

    var obj = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _classCallCheck(this, Schema);

    this.opts = opts;
    this.hooks = [];
    this.props = {};
    this.messages = Object.assign({}, Messages);
    this.validators = Object.assign({}, Validators);
    Object.keys(obj).forEach(function (k) {
      return _this.path(k, obj[k]);
    });
  }

  /**
   * Create or update `path` with given `rules`.
   *
   * @example
   * const schema = new Schema()
   * schema.path('name.first', { type: 'string' })
   * schema.path('name.last').type('string').required()
   *
   * @param {String} path - full path using dot-notation
   * @param {Object|Array|String|Schema|Property} [rules] - rules to apply
   * @return {Property}
   */

  _createClass(Schema, [{
    key: 'path',
    value: function path(_path, rules) {
      var _this2 = this;

      var parts = _path.split('.');
      var suffix = parts.pop();
      var prefix = parts.join('.');

      // Make sure full path is created
      if (prefix) {
        this.path(prefix);
      }

      // Array index placeholder
      if (suffix == '$') {
        this.path(prefix).type('array');
      }

      // Nested schema
      if (rules instanceof Schema) {
        rules.hook(function (k, v) {
          return _this2.path(join(k, _path), v);
        });
        return this.path(_path, rules.props);
      }

      // Return early when given a `Property`
      if (rules instanceof Property) {
        this.props[_path] = rules;
        // Notify parents if mounted
        this.propagate(_path, rules);
        return rules;
      }

      var prop = this.props[_path] || new Property(_path, this);

      this.props[_path] = prop;
      // Notify parents if mounted
      this.propagate(_path, prop);

      // No rules?
      if (!rules) return prop;

      // type shorthand
      // `{ name: 'string' }`
      if (typeof rules == 'string') {
        prop.type(rules);
        return prop;
      }

      // Allow arrays to be passed implicitly:
      // `{ keywords: [{ type: 'string' }]}`
      if (Array.isArray(rules)) {
        return this.path(join('$', _path), rules[0]);
      }

      var nested = false;

      // Check for nested objects
      for (var key in rules) {
        if (!rules.hasOwnProperty(key)) continue;
        if ('function' == typeof prop[key]) continue;
        nested = true;
        break;
      }

      Object.keys(rules).forEach(function (key) {
        var rule = rules[key];

        if (nested) {
          return _this2.path(join(key, _path), rule);
        }

        prop[key](rule);
      });

      return prop;
    }

    /**
     * Typecast given `obj`.
     *
     * @param {Object} obj - the object to typecast
     * @param {Object} [props]
     * @return {Object}
     * @private
     */

  }, {
    key: 'typecast',
    value: function typecast(obj) {
      var props = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this.generateProps(obj);

      // Count how many objects are typecasted
      var counter = 0;

      Object.keys(props).forEach(function (key) {
        var prop = props[key];
        var value = dot.get(obj, key);

        if (value == null) return;

        var cast = prop.typecast(value);

        if (cast === value) return;

        dot.set(obj, key, cast);
        counter++;
      });

      if (counter > 0) {
        // Make another pass in case there are generated array elements
        // E.g. '"1","2","3"' => ["1","2","3"] => [1, 2, 3]
        return this.typecast(obj);
      }

      return props;
    }

    /**
     * Strip all keys not defined in the schema
     *
     * @param {Object} obj - the object to strip
     * @param {Object} [props]
     * @param {String} [prefix]
     * @private
     */

  }, {
    key: 'strip',
    value: function strip(obj) {
      var _this3 = this;

      var props = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this.generateProps(obj);
      var prefix = arguments[2];

      Object.keys(obj).forEach(function (key) {
        var path = join(key, prefix);

        if (!props[path]) {
          // Delete is safe to use because all array indices should be
          // defined as paths in the props object
          delete obj[key];
          return;
        }

        if (['object', 'array'].includes(typeOf(obj[key]))) {
          _this3.strip(obj[key], props, path);
        }
      });
    }

    /**
     * Validate given `obj`.
     *
     * @example
     * const schema = new Schema({ name: { required: true }})
     * const errors = schema.validate({})
     * assert(errors.length == 1)
     * assert(errors[0].message == 'name is required')
     * assert(errors[0].path == 'name')
     *
     * @param {Object} obj - the object to validate
     * @param {Object} [opts] - options, see [Schema](#schema-1)
     * @return {Array}
     */

  }, {
    key: 'validate',
    value: function validate(obj) {
      var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      opts = Object.assign(this.opts, opts);

      var errors = [];

      // Generate props object and/or typecast
      // This object includes paths for all array elements.
      var props = opts.typecast ? this.typecast(obj) : this.generateProps(obj);

      if (opts.strip !== false) {
        // Pass in props to avoid regenerating
        this.strip(obj, props);
      }

      // Validate using the generated `props` object.
      Object.keys(props).forEach(function (path) {
        var value = dot.get(obj, path);
        var prop = props[path];
        var err = prop.validate(value, obj, path);
        if (err) errors.push(err);
      });

      return errors;
    }

    /**
     * Generate props for each valid path in the given object.
     *
     * @param {Object} obj
     * @return {Object}
     * @private
     */

  }, {
    key: 'generateProps',
    value: function generateProps(obj) {
      var _this4 = this;

      var props = {};

      Object.keys(this.props).forEach(function (key) {
        // Generate all valid paths, including array elements
        var paths = generatePaths(key, obj);
        // Aassign the original `Property` object to the generated paths.
        // E.g. some.path.1.in.an.array => some.path.$.in.an.array
        Object.keys(paths).forEach(function (p) {
          return props[p] = _this4.props[paths[p]];
        });
      });

      return props;
    }

    /**
     * Assert that given `obj` is valid.
     *
     * @example
     * const schema = new Schema({ name: 'string' })
     * schema.assert({ name: 1 }) // => Throws an error
     *
     * @param {Object} obj
     * @param {Object} [opts]
     */

  }, {
    key: 'assert',
    value: function assert(obj, opts) {
      var errors = this.validate(obj, opts);
      if (errors.length > 0) throw errors[0];
    }

    /**
     * Override default error messages.
     *
     * @example
     * const hex = (val) => /^0x[0-9a-f]+$/.test(val)
     * schema.path('some.path').use({ hex })
     * schema.message('hex', path => `${path} must be hexadecimal`)
     *
     * @example
     * schema.message({ hex: path => `${path} must be hexadecimal` })
     *
     * @param {String|Object} name - name of the validator or an object with name-message pairs
     * @param {String|Function} [message] - the message or message generator to use
     * @return {Schema}
     */

  }, {
    key: 'message',
    value: function message(name, _message) {
      var _this5 = this;

      if (typeof name == 'string') {
        this.messages[name] = _message;
        return this;
      }

      Object.keys(name).forEach(function (k) {
        return _this5.messages[k] = name[k];
      });
      return this;
    }

    /**
     * Override default validators.
     *
     * @example
     * schema.validator('required', val => val != null)
     *
     * @example
     * schema.validator({ required: val => val != null })
     *
     * @param {String|Object} name - name of the validator or an object with name-function pairs
     * @param {Function} [fn] - the function to use
     * @return {Schema}
     */

  }, {
    key: 'validator',
    value: function validator(name, fn) {
      var _this6 = this;

      if (typeof name == 'string') {
        this.validators[name] = fn;
        return this;
      }

      Object.keys(name).forEach(function (k) {
        return _this6.validators[k] = name[k];
      });
      return this;
    }

    /**
     * Accepts a function that is called whenever new props are added.
     *
     * @param {Function} fn - the function to call
     * @return {Schema}
     * @private
     */

  }, {
    key: 'hook',
    value: function hook(fn) {
      this.hooks.push(fn);
      return this;
    }

    /**
     * Notify all subscribers that a property has been added.
     *
     * @param {String} path - the path of the property
     * @param {Property} prop - the new property
     * @return {Schema}
     * @private
     */

  }, {
    key: 'propagate',
    value: function propagate(path, prop) {
      this.hooks.forEach(function (fn) {
        return fn(path, prop);
      });
      return this;
    }
  }]);

  return Schema;
}();

// Expose schema


module.exports = Schema;

/**
 * Join `path` with `prefix`
 *
 * @private
 */

function join(path, prefix) {
  return prefix ? prefix + '.' + path : path;
}

/**
 * Generate paths
 *
 * @private
 */

function generatePaths(path, obj) {
  var prefix = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '';
  var map = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

  var prev = map[prefix] || '';

  // Return early if there are no arrays
  if (!path.includes('.$')) {
    map[prefix + path] = prev + path;
    return map;
  }

  // Split by arrays
  var parts = path.split('.$');
  var first = parts.shift();
  var arr = dot.get(obj, prefix + first);

  if (!Array.isArray(arr)) {
    return map;
  }

  arr.forEach(function (val, i) {
    var newPath = join(i, prefix + first);
    map[newPath] = join('$', prev + first);
    generatePaths(parts.join('.$'), obj, newPath, map);
  });

  return map;
}