'use strict';

var typeOf = require('component-type');

/**
 * Default validators.
 *
 * @private
 */

var Validators = {
  /**
   * Validates presence.
   *
   * @param {Mixed} value - the value being validated
   * @param {Object} ctx - the object being validated
   * @param {Bolean} required
   * @return {Boolean}
   * @api public
   */

  required: function required(value, ctx, _required) {
    if (_required === false) return true;
    return value != null && value !== '';
  },


  /**
   * Validates type.
   *
   * @param {Mixed} value - the value being validated
   * @param {Object} ctx - the object being validated
   * @param {String} name name of the type
   * @return {Boolean}
   * @api public
   */

  type: function type(value, ctx, name) {
    return typeOf(value) == name;
  },


  /**
   * Validates length.
   *
   * @param {String} value the string being validated
   * @param {Object} ctx the object being validated
   * @param {Object} rules object with .min and/or .max props.
   * @param {Number} [rules.min] - minimum length
   * @param {Number} [rules.max] - maximum length
   * @return {Boolean}
   * @api public
   */

  length: function length(value, ctx, rules) {
    var min = rules.min,
        max = rules.max;

    if (min && value.length < min) return false;
    if (max && value.length > max) return false;
    return true;
  },


  /**
   * Validates enums.
   *
   * @param {String} value the string being validated
   * @param {Object} ctx the object being validated
   * @param {Array} enums array with allowed values
   * @return {Boolean}
   * @api public
   */

  enum: function _enum(value, ctx, enums) {
    return enums.includes(value);
  },


  /**
   * Validates against given `regexp`.
   *
   * @param {String} value the string beign validated
   * @param {Object} ctx the object being validated
   * @param {RegExp} regexp the regexp to validate against
   * @return {Boolean}
   * @api public
   */

  match: function match(value, ctx, regexp) {
    return regexp.test(value);
  }
};

module.exports = Validators;