'use strict';

var Property = require('../lib/property');
var Schema = require('../lib/schema');
var messages = require('../lib/messages');

describe('Property', function () {
  it('should have a .name property', function () {
    var prop = new Property('test', new Schema());
    prop.name.should.equal('test');
  });

  describe('.use()', function () {
    it('should register each object property as a validator', function () {
      var prop = new Property('test', new Schema());
      prop.use({
        one: function one(v) {
          return v != 1;
        },
        two: function two(v) {
          return v != 2;
        }
      });
      prop.validate(1).should.be.an.instanceOf(Error);
      prop.validate(2).should.be.an.instanceOf(Error);
      prop.validate(3).should.equal(false);
    });

    it('should use property names to look up error messages', function () {
      var schema = new Schema();
      var prop = new Property('test', schema);

      schema.message({
        one: function one() {
          return 'error 1';
        },
        two: function two() {
          return 'error 2';
        }
      });

      prop.use({
        one: function one(v) {
          return v != 1;
        },
        two: function two(v) {
          return v != 2;
        }
      });

      prop.validate(1).message.should.equal('error 1');
      prop.validate(2).message.should.equal('error 2');
    });

    it('should register additional arguments', function () {
      var prop = new Property('test', new Schema());
      var first = void 0,
          second = void 0;
      prop.use({
        one: [function (v, c, arg) {
          return first = arg;
        }, 1],
        two: [function (v, c, arg) {
          return second = arg;
        }, 2]
      });
      prop.validate({ test: 1 });
      first.should.equal(1);
      second.should.equal(2);
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.use({ one: function one() {} }).should.equal(prop);
    });
  });

  describe('.required()', function () {
    it('should register a validator', function () {
      var prop = new Property('test', new Schema());
      prop.required();
      prop.validate(null).should.be.an.instanceOf(Error);
      prop.validate(100).should.equal(false);
    });

    it('should use the correct error message', function () {
      var prop = new Property('test', new Schema());
      var message = messages.required(prop.name, {}, true);
      prop.required();
      prop.validate(null).message.should.equal(message);
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.required().should.equal(prop);
    });
  });

  describe('.type()', function () {
    it('should register a validator', function () {
      var prop = new Property('test', new Schema());
      prop.type('string');
      prop.validate(1).should.be.an.instanceOf(Error);
      prop.validate('test').should.equal(false);
      prop.validate(null).should.equal(false);
    });

    it('should set the internal ._type property', function () {
      var prop = new Property('test', new Schema());
      prop.type('string');
      prop._type.should.equal('string');
    });

    it('should use the correct error message', function () {
      var prop = new Property('test', new Schema());
      var message = messages.type(prop.name, {}, 'string');
      prop.type('string');
      prop.validate(1).message.should.equal(message);
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.type('number').should.equal(prop);
    });
  });

  describe('.match()', function () {
    it('should register a validator', function () {
      var prop = new Property('test', new Schema());
      prop.match(/^abc$/);
      prop.validate('cab').should.be.an.instanceOf(Error);
      prop.validate('abc').should.equal(false);
      prop.validate(null).should.equal(false);
    });

    it('should use the correct error message', function () {
      var prop = new Property('test', new Schema());
      var regexp = /^abc$/;
      var message = messages.match(prop.name, {}, regexp);
      prop.match(regexp);
      prop.validate('cab').message.should.equal(message);
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.match(/abc/).should.equal(prop);
    });
  });

  describe('.length()', function () {
    it('should register a validator', function () {
      var prop = new Property('test', new Schema());
      prop.length({ min: 2, max: 3 });
      prop.validate('abcd').should.be.an.instanceOf(Error);
      prop.validate('a').should.be.an.instanceOf(Error);
      prop.validate('abc').should.equal(false);
      prop.validate(null).should.equal(false);
    });

    it('should use the correct error message', function () {
      var prop = new Property('test', new Schema());
      var rule = { max: 1 };
      var message = messages.length(prop.name, {}, rule);
      prop.length(rule);
      prop.validate('abc').message.should.equal(message);
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.length({}).should.equal(prop);
    });
  });

  describe('.enum()', function () {
    it('should register a validator', function () {
      var prop = new Property('test', new Schema());
      prop.enum(['one', 'two']);
      prop.validate('three').should.be.an.instanceOf(Error);
      prop.validate('one').should.equal(false);
      prop.validate('two').should.equal(false);
      prop.validate(null).should.equal(false);
    });

    it('should use the correct error message', function () {
      var prop = new Property('test', new Schema());
      var enums = ['one', 'two'];
      var message = messages.enum(prop.name, {}, enums);
      prop.enum(enums);
      prop.validate('three').message.should.equal(message);
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.enum(['one', 'two']).should.equal(prop);
    });
  });

  describe('.schema()', function () {
    it('should mount given schema on parent schema ', function () {
      var schema1 = new Schema();
      var schema2 = new Schema({ world: { required: true } });
      var prop1 = schema1.path('hello');
      var prop2 = schema2.path('world');
      prop1.schema(schema2);
      schema1.path('hello.world').should.equal(prop2);
    });

    it('should support chaining', function () {
      var schema = new Schema();
      var prop = new Property('test', new Schema());
      prop.schema(schema).should.equal(prop);
    });
  });

  describe('.each()', function () {
    it('should define a new array path on the parent schema', function () {
      var schema = new Schema();
      var prop = new Property('test', schema);
      prop.each({ type: 'number' });
      schema.path('test.$')._type.should.equal('number');
    });

    it('should support chaining', function () {
      var prop = new Property('test', new Schema());
      prop.each({}).should.equal(prop);
    });
  });

  describe('.typecast()', function () {
    it('should typecast given value to the type defined by ._type', function () {
      var prop = new Property('test', new Schema());
      prop.type('string');
      prop.typecast(123).should.equal('123');
    });
  });

  describe('.validate()', function () {
    it('should validate using using registered validators', function () {
      var prop = new Property('test', new Schema());
      prop.required();
      prop.match(/^abc$/);
      prop.validate(null).should.be.an.instanceOf(Error);
      prop.validate('abc').should.equal(false);
      prop.validate('cab').should.be.an.instanceOf(Error);
    });

    it('should run `required` and `type` validators first', function () {
      var schema = new Schema();
      var prop = new Property('test', schema);
      var done = {};
      var counter = 0;

      schema.validator({
        required: function required() {
          return done.required = ++counter;
        },
        type: function type() {
          return done.type = ++counter;
        },
        match: function match() {
          return done.match = ++counter;
        },
        enum: function _enum() {
          return done.enum = ++counter;
        }
      });

      prop.match();
      prop.enum();
      prop.type();
      prop.required();
      prop.validate('something');

      done.required.should.eql(1);
      done.type.should.eql(2);
    });

    it('should assign errors a .path', function () {
      var prop = new Property('some.path', new Schema());
      prop.required();
      prop.validate(null).path.should.equal('some.path');
    });

    it('should allow path to be overridden', function () {
      var prop = new Property('some.path', new Schema());
      prop.required();
      prop.validate(null, {}, 'some.other.path').path.should.equal('some.other.path');
    });

    it('should pass context to validators', function () {
      var prop = new Property('test', new Schema());
      var obj = { hello: 'world' };
      var ctx = void 0;

      prop.use({
        context: function context(v, c) {
          ctx = c;
        }
      });

      prop.validate('abc', obj);
      obj.should.equal(ctx);
    });
  });

  describe('.path()', function () {
    it('should proxy all arguments to parent schema', function () {
      var schema = new Schema();
      var prop = new Property('test', schema);
      var ret = prop.path('hello');
      schema.path('hello').should.equal(ret);
    });
  });
});