'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var Schema = require('../lib/schema');
var Property = require('../lib/property');

describe('Schema', function () {
  context('when given an object', function () {
    it('should create properties', function () {
      var schema = new Schema({ a: { type: 'string' } });
      schema.props.should.have.property('a');
    });
  });

  describe('.path()', function () {
    it('should create properties', function () {
      var schema = new Schema();
      schema.path('a', { type: 'string' });
      schema.props.should.have.property('a');
    });

    it('should allow type shorthand', function () {
      var schema = new Schema();
      schema.path('a', 'string');
      schema.props['a']._type.should.equal('string');
    });

    it('should create properties for all subpaths', function () {
      var schema = new Schema();
      schema.path('hello.planet.earth');
      schema.props.should.have.property('hello');
      schema.props.should.have.property('hello.planet');
      schema.props.should.have.property('hello.planet.earth');
    });

    it('should support nested properties', function () {
      var schema = new Schema();
      schema.path('a', { b: { type: 'string' } });
      schema.props.should.have.property('a.b');
    });

    it('should register validators', function () {
      var schema = new Schema();
      schema.path('a', { b: { required: true } });
      schema.validate({}).should.have.length(1);
    });

    it('should return a Property', function () {
      var schema = new Schema();
      schema.path('a').should.be.instanceOf(Property).and.have.property('name', 'a');
    });

    it('should work with nested schemas', function () {
      var schema1 = new Schema();
      var schema2 = new Schema();
      schema2.path('hello', { required: true });
      schema1.path('schema2', schema2).required(true);
      schema1.props.should.have.property('schema2.hello');
      schema1.validate({}).should.have.length(2);
      schema1.validate({ schema2: { hello: null } }).should.have.length(1);
      schema1.validate({ schema2: { hello: 'world' } }).should.have.length(0);
    });

    it('should propagate new props from nested schema', function () {
      var schema1 = new Schema();
      var schema2 = new Schema();
      schema1.path('schema2', schema2);
      schema2.path('hello', { required: true });
      schema2.path('hello.world', { required: true });
      schema1.props.should.have.property('schema2.hello');
      schema1.props.should.have.property('schema2.hello.world');
    });

    context('when given a path ending with $', function () {
      it('should set `property.type` to array', function () {
        var schema = new Schema();
        schema.path('hello.$');
        schema.props.hello._type.should.equal('array');
      });

      it('should apply rules to each element in the array', function () {
        var schema = new Schema();
        schema.path('hello.$').type('number');
        schema.props.hello._type.should.equal('array');
        schema.props['hello.$']._type.should.equal('number');
      });
    });
  });

  describe('.strip()', function () {
    it('should delete all keys not in the schema', function () {
      var schema = new Schema({
        a: { type: 'number' },
        b: [{ a: { type: 'number' } }],
        c: { a: { type: 'number' } }
      });

      var obj = { a: 1, b: [{ a: 1, b: 1 }, { a: 1 }], c: { a: 1, b: 1 }, d: 1 };
      schema.strip(obj);
      obj.should.deepEqual({ a: 1, b: [{ a: 1 }, { a: 1 }], c: { a: 1 } });
    });
  });

  describe('.validate()', function () {
    it('should return an array of errors', function () {
      var schema = new Schema({ name: { type: 'string' } });
      var res = schema.validate({ name: 123 });
      res.should.be.an.Array();
      res.should.have.length(1);
    });

    it('should set the correct paths on the error objects', function () {
      var schema = new Schema({ things: [{ type: 'string' }] });
      var res = schema.validate({ things: ['car', 1, 3] });

      var _res = _slicedToArray(res, 2),
          err1 = _res[0],
          err2 = _res[1];

      res.should.have.length(2);
      err1.path.should.equal('things.1');
      err1.message.should.match(/things\.1/);
      err2.path.should.equal('things.2');
      err2.message.should.match(/things\.2/);
    });

    it('should work with $ a placeholder for array indices', function () {
      var schema = new Schema();
      schema.path('a.$.b').required();
      schema.path('a.$.b.$').type('string');
      schema.path('a.$.c.$.$').type('string');
      var res = schema.validate({
        a: [{ b: ['hello', 'world'] }, { b: ['hello', 1] }, { c: [['hello', 'world'], ['hello', 2]] }]
      });
      res.should.have.length(3);
    });

    it('should strip by default', function () {
      var schema = new Schema({ a: { type: 'number' } });
      var obj = { a: 1, b: 1 };
      var res = schema.validate(obj);
      obj.should.deepEqual({ a: 1 });
    });

    context('with strip disabled', function () {
      it('should not delete any keys', function () {
        var obj = { name: 'name', age: 23 };
        var schema = new Schema({ name: { type: 'string' } });
        var res = schema.validate(obj, { strip: false });
        obj.should.have.property('age', 23);
      });
    });

    context('with typecasting enabled', function () {
      it('should typecast before validation', function () {
        var schema = new Schema({ name: { type: 'string' } });
        var res = schema.validate({ name: 123 }, { typecast: true });
        res.should.have.length(0);
      });

      it('should typecast arrays and elements within arrays', function () {
        var schema = new Schema();
        schema.path('a.$.b').required();
        schema.path('a.$.b.$').type('string');
        schema.path('a.$.c.$.$').type('string');
        schema.path('b.$').type('number');

        var obj = {
          a: [{ b: ['a', 'b'] }, { b: 1, c: [['a', 'b'], ['a', 2]] }],
          b: '1,2,3,4,5'
        };

        var res = schema.validate(obj, { typecast: true });
        res.should.have.length(0);

        obj.should.deepEqual({
          a: [{ b: ['a', 'b'] }, { b: ['1'], c: [['a', 'b'], ['a', '2']] }],
          b: [1, 2, 3, 4, 5]
        });
      });

      it('should not typecast undefined', function () {
        var schema = new Schema({ name: { type: 'string' } });
        var wrap = function wrap() {
          return schema.validate({}, { typecast: true });
        };
        wrap.should.not.throw();
      });
    });
  });

  describe('.assert()', function () {
    it('should throw if validation fails', function () {
      var schema = new Schema({ name: { type: 'string' } });
      var wrap = function wrap() {
        return schema.assert({ name: 123 });
      };
      wrap.should.throw();
    });
  });

  describe('.message()', function () {
    it('should set default messages', function () {
      var schema = new Schema({ name: { required: true } });
      schema.message('required', 'test');

      var _schema$validate = schema.validate({}),
          _schema$validate2 = _slicedToArray(_schema$validate, 1),
          error = _schema$validate2[0];

      error.message.should.equal('test');
    });

    it('should accept an object of name-message pairs', function () {
      var schema = new Schema({ name: { required: true } });
      schema.message({ required: 'test' });

      var _schema$validate3 = schema.validate({}),
          _schema$validate4 = _slicedToArray(_schema$validate3, 1),
          error = _schema$validate4[0];

      error.message.should.equal('test');
    });
  });

  describe('.validator()', function () {
    it('should set default validators', function () {
      var schema = new Schema({ name: { required: true } });
      schema.validator('required', function () {
        return false;
      });

      var _schema$validate5 = schema.validate({ name: 'hello' }),
          _schema$validate6 = _slicedToArray(_schema$validate5, 1),
          error = _schema$validate6[0];

      error.message.should.equal('name is required.');
    });

    it('should accept an object of name-function pairs', function () {
      var schema = new Schema({ name: { required: true } });
      schema.validator({ required: function required() {
          return false;
        } });

      var _schema$validate7 = schema.validate({ name: 'hello' }),
          _schema$validate8 = _slicedToArray(_schema$validate7, 1),
          error = _schema$validate8[0];

      error.message.should.equal('name is required.');
    });
  });
});